/*
 * $Id: w25q_flash.h 499 2021-07-26 05:24:02Z true $
 * begin 20210711 true
 * 
 * winbond w25qxx driver, standard SPI mode
 * 
 */

#ifndef INC_DRIVER_W25Q_FLASH_H_
#define INC_DRIVER_W25Q_FLASH_H_


#include "gd32vf103.h"
#include "driver/gd32v_hw_spi.h"



#define W25Q_SPI_DEV				SPI0
#define W25Q_SPI_CS_PORT			GPIOC
#define W25Q_SPI_CS_PIN				GPIO_PIN_15

#define W25Q_SPI_XFER(r, t, l)		spi_xfer(W25Q_SPI_DEV, r, t, l & 0xffff)
#define W25Q_CS()					gpio_bit_reset(W25Q_SPI_CS_PORT, W25Q_SPI_CS_PIN)
#define W25Q_NOCS()					gpio_bit_set(W25Q_SPI_CS_PORT, W25Q_SPI_CS_PIN)


// registers
#define W25Q_SR1_BUSY				0x01
#define W25Q_SR1_WEL				0x02
#define W25Q_SR1_BP0				0x04
#define W25Q_SR1_BP1				0x08
#define W25Q_SR1_BP2				0x10
#define W25Q_SR1_TB					0x20
#define W25Q_SR1_SEC				0x40
#define W25Q_SR1_SRP0				0x80

#define W25Q_SR2_SRP1				0x01
#define W25Q_SR2_QE					0x02

#define W25Q_BP_MASK				0x1c

// table 1, parenthesis are read data
#define W25Q_I_WRITE_ENABLE         0x06	//								sets WEL
#define W25Q_I_WRITE_DISABLE        0x04	//								clears WEL
#define W25Q_I_READ_STATUS_1		0x05    // 1: (low byte)
#define W25Q_I_READ_STATUS_2		0x35    // 1: (high byte)
#define W25Q_I_WRITE_STATUS			0x01    // 2: low byte, high byte
#define W25Q_I_PAGE_PROGRAM			0x02    // 4: addr[2:0], data byte, ...	requires WEL set
#define W25Q_I_QUAD_PAGE_PROGRAM	0x32    // unimplemented
#define W25Q_I_BLOCK_ERASE_64K		0xd8    // 3: addr[2:0]					requires WEL set
#define W25Q_I_BLOCK_ERASE_32K		0x52    // 3: addr[2:0]					requires WEL set
#define W25Q_I_SECTOR_ERASE_4K		0x20    // 3: addr[2:0]					requires WEL set
#define W25Q_I_CHIP_ERASE           0xc7	//								requires WEL set
#define W25Q_I_ERASE_SUSPEND		0x75
#define W25Q_I_ERASE_RESUME			0x7a
#define W25Q_I_POWER_DOWN			0xb9
#define W25Q_I_HIGH_PERF_MODE		0xa3	// 3: dummy bytes
#define W25Q_I_MODE_BIT_RESET		0xff	// 1: 0xff
#define W25Q_I_POWER_UP				0xab	// 4: dummy[2:0], (id low byte)
#define W25Q_I_MFR_DEV_ID			0x90	// 5: dummy[1:0], 0x00, (mfr byte, id low byte)
#define W25Q_I_UNIQUE_ID			0x4b	// 5: dummy[3:0], (uid[12:4])
#define W25Q_I_JEDEC_ID				0x9f	// 3: (mfr byte, id high byte, id low byte)

// table 2 (reading)
#define W25Q_I_READ					0x03	// 4: addr[2:0], (data byte), ...
#define W25Q_I_FAST_READ			0x03	// 5: addr[2:0], dummy, (data byte), ...



int8_t w25q_read(uint32_t addr, uint8_t *dat, uint32_t bytes);
int8_t w25q_write(uint32_t addr, uint8_t *dat, uint32_t bytes);
int8_t w25q_erase_sector(uint32_t addr, uint8_t blocking);

void w25q_power_up();
void w25q_power_down();

uint8_t  w25q_get_sr1();
uint8_t  w25q_get_sr2();
uint16_t w25q_get_status();

uint32_t w25q_get_jedec();

#endif