/*
 * $Id: btn.h 500 2021-08-08 19:43:38Z true $
 * begin 20190527 true
 *
 * button ticks are 1/1024 of a second
 */


#ifndef INC_UI_BTN_H_
#define INC_UI_BTN_H_



#include "gd32vf103.h"
#include <stdint.h>



typedef struct BtnData {
	uint8_t state;			// what is the current handled state of this button
	uint8_t hold_time;		// hold time in 1/16 second increments
	uint16_t hold_retrig;	// hold retrigger in button ticks, 0 disables
	uint16_t hold_rt_ctr;	// hold retrigger counter
	uint16_t held;			// actual hold counts in button ticks
	void (*push_cb)();
	void (*held_cb)();
	void (*release_cb)();
} BtnData;



#define BTN_COUNT			4

#define BTN_USART			USART1
#define BTN_USART_IRQn		USART1_IRQn
#define BTN_USART_RCU		RCU_USART1

#define BTN_USART_IRQ_LEVEL	14

#define BTN_PUSH			(1 << 0)
#define BTN_HELD			(1 << 1)
#define BTN_RELEASE			(1 << 2)

#define BTN_PUSH_CB			(1 << 4)
#define BTN_HELD_CB			(1 << 5)
#define BTN_RELEASE_CB		(1 << 6)


#define BTN_DEBOUNCE		1		// how many button ticks to wait before registering press
#define BTN_MAX_HOLD		30000	// longest reported / processed hold time (30s)
#define BTN_HOLD_SHIFT		4		// rshift value to get ~1/32 sec hold time


#define BTN_HOLDTIME_0_25S	8
#define BTN_HOLDTIME_0_50S	16
#define BTN_HOLDTIME_1_00S	31
#define BTN_HOLDTIME_2_00S	(31 * 2)
#define BTN_HOLDTIME_3_00S	(31 * 3)
#define BTN_HOLDTIME_4_00S	(31 * 4)
#define BTN_HOLDTIME_5_00S	(31 * 5)

#define BTN_DATA_VALID		(btn_data[0] == 0x1d && btn_data[1] == 0x29)
#define BTN_DATA_SN_BYTE	btn_data[5]



extern BtnData btn[BTN_COUNT];
extern uint8_t btn_data[7];
extern uint32_t btn_invalid_packet_cnt;
extern uint32_t btn_short_packet_cnt;



void btn_init();
void btn_tick();
void btn_callback();

void btn_data_isr(uint8_t flag);



#endif /* INC_UI_BTN_H_ */
