/*
 * $Id: gd32v_hw_adc.c 498 2021-07-23 19:38:20Z true $
 * begin 20210621 by true
 * 
 * adc + dma (maybe?) routines
 */

#include "driver/gd32v_hw_adc.h"


volatile uint16_t adc_data_raw[ADC_CHAN_COUNT];
volatile uint16_t adc_data_avg[ADC_CHAN_COUNT];
volatile uint16_t adc_data_avg_sample[ADC_CHAN_COUNT][8];
uint8_t adc_idx = 0;


static dma_parameter_struct dma_param_ch0;


void gd32v_adc_init()
{
    adc_deinit(ADC0);

	rcu_adc_clock_config(RCU_CKADC_CKAPB2_DIV4);

    // adc general configuration
	adc_mode_config(ADC_MODE_FREE);
	adc_special_function_config(ADC0, ADC_SCAN_MODE, ENABLE);
	adc_data_alignment_config(ADC0, ADC_DATAALIGN_RIGHT);

    // enable temperature sensor
    adc_tempsensor_vrefint_enable();
	
	// add channels used in firmware to scan list
	adc_channel_length_config(ADC0, ADC_REGULAR_CHANNEL, ADC_CHAN_COUNT);
	adc_regular_channel_config(ADC0, 0, ADC_CHANNEL_1,  ADC_SAMPLETIME_41POINT5);		// lightsensor led
    adc_regular_channel_config(ADC0, 1, ADC_CHANNEL_4,  ADC_SAMPLETIME_71POINT5);		// battery voltage
	adc_regular_channel_config(ADC0, 2, ADC_CHANNEL_16, ADC_SAMPLETIME_239POINT5);		// temperature
    adc_regular_channel_config(ADC0, 3, ADC_CHANNEL_17, ADC_SAMPLETIME_239POINT5);		// bandgap (for battery comparison)

	// we're software triggering the regular channel
    adc_external_trigger_source_config(ADC0, ADC_REGULAR_CHANNEL, ADC0_1_EXTTRIG_REGULAR_NONE);
    adc_external_trigger_config(ADC0, ADC_REGULAR_CHANNEL, ENABLE);
    
    // dma parameters
    dma_deinit(DMA0, DMA_CH0);
    dma_param_ch0.periph_addr  = (uint32_t)(&ADC_RDATA(ADC0));
    dma_param_ch0.periph_inc   = DMA_PERIPH_INCREASE_DISABLE;
    dma_param_ch0.periph_width = DMA_PERIPHERAL_WIDTH_16BIT;
    dma_param_ch0.memory_addr  = (uint32_t)(adc_data_raw);
    dma_param_ch0.memory_inc   = DMA_MEMORY_INCREASE_ENABLE;
    dma_param_ch0.memory_width = DMA_MEMORY_WIDTH_16BIT;
    dma_param_ch0.direction    = DMA_PERIPHERAL_TO_MEMORY;
    dma_param_ch0.number       = ADC_CHAN_COUNT;
    dma_param_ch0.priority     = DMA_PRIORITY_HIGH;
    dma_init(DMA0, DMA_CH0, &dma_param_ch0);

    // use circular mode so we don't need to set up every time
    dma_circulation_enable(DMA0, DMA_CH0);

    // enable DMA interrupt to let us know when it's finished
    eclic_irq_enable(DMA0_Channel0_IRQn, 0, 0);
    dma_interrupt_flag_clear(DMA0, DMA_CH0, DMA_INT_FLAG_G);
	dma_interrupt_enable(DMA0, DMA_CH0, DMA_INT_FTF);

    // enable DMA
    adc_dma_mode_enable(ADC0);
    dma_channel_enable(DMA0, DMA_CH0);

    adc_enable(ADC0);
}

static void gd32v_adc_convert_start()
{
    adc_software_trigger_enable(ADC0, ADC_REGULAR_CHANNEL);
}

void gd32v_adc_tick()
{
    if (!(ADC_CTL1(ADC0) & ADC_CTL1_CLB)) {
        gd32v_adc_convert_start();
    }
}

void gd32v_adc_cal_temp(uint16_t offset_deg_c_x10)
{
    

}

uint16_t adc_get_temp_degc_x10()
{
    uint16_t t;

    // todo: verify / fix this function, not sure it actually works as expected
    t = (((1420 - (adc_data_avg[ADC_CHAN_TEMPERATURE] * 3300) / 4096) * 10000) / 4300) + 115;

    return t;
}

void gd32v_adc_dma_complete_cb()
{
    int i, j;
    uint32_t s;

    for (i = 0; i < ADC_CHAN_COUNT; i++) {
        adc_data_avg_sample[i][adc_idx] = adc_data_raw[i];
    }

    adc_idx++;
    adc_idx %= 8;

    // average after next 8 arrive
    if (!adc_idx) {
        for (i = 0; i < ADC_CHAN_COUNT; i++) {
            s = 0;
            for (j = 0; j < 8; j++) s += adc_data_avg_sample[i][j];
            adc_data_avg[i] = s >> 3;
        }
    }

    // if we're not calibrating, do next conversion
    /*
    if (!(ADC_CTL1(ADC0) & ADC_CTL1_CLB)) {
        gd32v_adc_convert_start();
    }
    */
}

uint16_t adc_get_val_raw(uint8_t chan)
{
    if (chan < ADC_CHAN_COUNT) {
        return adc_data_raw[chan];
    }
    return 0xffff;
}

uint16_t adc_get_val_avg(uint8_t chan)
{
    if (chan < ADC_CHAN_COUNT) {
        return adc_data_avg[chan];
    }
    return 0xffff;
}