/*
 * $Id: gd32v_hw_clk.c 497 2021-07-23 03:42:57Z true $
 * begin 20210617 true
 * 
 * clock set / sleep routines
 */

#include "driver/gd32v_hw_clk.h"


void system_clock_set_irc8m(uint32_t pllmul)
{
    uint32_t timeout = 0U;
    uint32_t stab_flag = 0U;

    // enable IRC8M
    RCU_CTL |= RCU_CTL_IRC8MEN;
    rcu_system_clock_source_config(RCU_CKSYSSRC_IRC8M);
    rcu_deinit();
    
    // wait until IRC8M is stable or the startup time is longer than IRC8M_STARTUP_TIMEOUT
    do {
        timeout++;
        stab_flag = (RCU_CTL & RCU_CTL_IRC8MSTB);
    } while(!stab_flag && (IRC8M_STARTUP_TIMEOUT != timeout));

    // failed?
    if(!(RCU_CTL & RCU_CTL_IRC8MSTB)){
        // welp we done. if we don't have clocks we don't have a system
        while (1);
    }

    // ok, IRC8M is stable, so set bus clocks
    RCU_CFG0 |= RCU_AHB_CKSYS_DIV1;     // AHB = SYSCLK
    RCU_CFG0 |= RCU_APB2_CKAHB_DIV1;    // APB2 = AHB/1
    RCU_CFG0 |= RCU_APB1_CKAHB_DIV1;    // APB1 = AHB/2

    // CK_PLL = (CK_IRC8M/2) * MUL = CK_SYS
    RCU_CFG0 &= ~(RCU_CFG0_PLLMF | RCU_CFG0_PLLMF_4);
    RCU_CFG0 |= pllmul;     // RCU_PLL_MUL27

    // enable PLL
    RCU_CTL |= RCU_CTL_PLLEN;

    // wait until PLL is stable
    while (!(RCU_CTL & RCU_CTL_PLLSTB));

    // select PLL as system clock
    RCU_CFG0 &= ~RCU_CFG0_SCS;
    RCU_CFG0 |= RCU_CKSYSSRC_PLL;

    // wait until PLL is selected as system clock
    while(!(RCU_CFG0 & RCU_SCSS_PLL));
}

void enable_rcu_clocks()
{
    // obvious I hope
	rcu_periph_clock_enable(RCU_GPIOA);
    rcu_periph_clock_enable(RCU_GPIOB);
    rcu_periph_clock_enable(RCU_GPIOC);
    
    // dma is used with i2c and maybe other things in the future
    rcu_periph_clock_enable(RCU_DMA0);
}