/*
 * $Id: gd32v_hw_spi.c 494 2021-07-21 11:46:11Z true $
 * begin 20210709
 *
 * polling mode SPI
 */

#include "driver/gd32v_hw_spi.h"


void gd32v_spi_init(uint32_t dev, uint8_t prescale)
{
    spi_parameter_struct spi_init_struct;
    
    // deinitialize SPI and the parameters
    spi_i2s_deinit(SPI0);
    spi_struct_para_init(&spi_init_struct);

    // configure SPI0 parameters
    spi_init_struct.trans_mode           = SPI_TRANSMODE_FULLDUPLEX;
    spi_init_struct.device_mode          = SPI_MASTER;
    spi_init_struct.frame_size           = SPI_FRAMESIZE_8BIT;
    spi_init_struct.clock_polarity_phase = SPI_CK_PL_HIGH_PH_2EDGE;
    spi_init_struct.nss                  = SPI_NSS_SOFT;
    spi_init_struct.prescale             = prescale;
    spi_init_struct.endian               = SPI_ENDIAN_MSB;
    spi_init(SPI0, &spi_init_struct);
}

uint32_t gd32v_spi_get_prescale(uint32_t dev)
{
    return (SPI_CTL0(dev) & SPI_CTL0_PSC) >> 3;
}

void gd32v_spi_set_prescale(uint32_t dev, uint32_t prescale)
{
    SPI_CTL0(dev) = (SPI_CTL0(dev) & ~(SPI_CTL0_PSC)) | CTL0_PSC(prescale);
}

void gd32v_spi_set_polarity(uint32_t dev, uint8_t polarity)
{
    SPI_CTL0(dev) = (SPI_CTL0(dev) & ~(SPI_CTL0_CKPH)) | (polarity ? (SPI_CTL0_CKPH) : 0);
}

uint8_t spi_xfer_byte(uint32_t dev, uint8_t byte)
{
    // send byte
    while(!spi_i2s_flag_get(dev, SPI_FLAG_TBE));
    spi_i2s_data_transmit(dev, byte);
    // receive byte
    while(!spi_i2s_flag_get(dev, SPI_FLAG_RBNE));
    return spi_i2s_data_receive(dev);
}

void spi_xfer(uint32_t dev, uint8_t *rx, const uint8_t *tx, uint16_t len)
{
    uint16_t i;
	uint8_t r;

	// make sure you selected your device before proceeding!
    spi_enable(dev);

	i = 0;
	while (i < len) {
		r = spi_xfer_byte(dev, tx ? tx[i] : 0x00);
        if (rx) { rx[i] = r; }
		i++;
	}

	// make sure you deselect your device when done!
    spi_disable(dev);
}
