/*
 * $Id: gd32v_hw_tim.c 499 2021-07-26 05:24:02Z true $
 * begin 20210620 true
 * 
 * timer init and interrupt callbacks
 */

#include "driver/gd32v_hw_tim.h"



timer_oc_parameter_struct timer_oc_para;
timer_parameter_struct    timer_para;


/* 1khz mainline timer */
void tim_rgb_init()
{
	rcu_periph_clock_enable(RCU_TIMER0);
	rcu_periph_clock_enable(RCU_TIMER2);
	rcu_periph_clock_enable(RCU_TIMER3);

    // default structs
    timer_struct_para_init(&timer_para);
    timer_channel_output_struct_para_init(&timer_oc_para);

    // eyes are timer0, which is 1/2 duty cycle
    // configure for 10 bit operation
    timer_deinit(TIMER0);
    timer_para.period          = 1023;
    timer_init(TIMER0, &timer_para);
    
    timer_auto_reload_shadow_enable(TIMER0);
    
    timer_oc_para.outputstate  = TIMER_CCX_ENABLE;
    timer_channel_output_config(TIMER0, TIMER_CH_0, &timer_oc_para);
    timer_channel_output_config(TIMER0, TIMER_CH_1, &timer_oc_para);
    timer_channel_output_config(TIMER0, TIMER_CH_2, &timer_oc_para);

    TIMER_CH0CV(TIMER0) = 0;
    TIMER_CH1CV(TIMER0) = 0;
    TIMER_CH2CV(TIMER0) = 0;

    timer_channel_output_mode_config  (TIMER0, TIMER_CH_0, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER0, TIMER_CH_0, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER0, TIMER_CH_1, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER0, TIMER_CH_1, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER0, TIMER_CH_2, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER0, TIMER_CH_2, TIMER_OC_SHADOW_DISABLE);

    // edges are timer2 and timer3, and are 1/5 duty cycle
    // configure for 10 bit operation (eventually)
    timer_deinit(TIMER2);
    timer_deinit(TIMER3);
    timer_para.period          = 255;
    timer_init(TIMER2, &timer_para);
    timer_init(TIMER3, &timer_para);

    timer_auto_reload_shadow_enable(TIMER2);
    timer_auto_reload_shadow_enable(TIMER3);

    timer_oc_para.ocpolarity   = TIMER_OC_POLARITY_LOW;
    timer_oc_para.ocidlestate  = TIMER_OC_IDLE_STATE_HIGH;
    timer_channel_output_config(TIMER2, TIMER_CH_0, &timer_oc_para);
    timer_channel_output_config(TIMER2, TIMER_CH_1, &timer_oc_para);
    timer_channel_output_config(TIMER2, TIMER_CH_2, &timer_oc_para);
    timer_channel_output_config(TIMER2, TIMER_CH_3, &timer_oc_para);

    timer_channel_output_config(TIMER3, TIMER_CH_2, &timer_oc_para);
    timer_channel_output_config(TIMER3, TIMER_CH_3, &timer_oc_para);

    TIMER_CH0CV(TIMER2) = 0;
    TIMER_CH1CV(TIMER2) = 0;
    TIMER_CH2CV(TIMER2) = 0;
    TIMER_CH3CV(TIMER2) = 0;

    TIMER_CH2CV(TIMER3) = 0;
    TIMER_CH3CV(TIMER3) = 0;

    timer_channel_output_mode_config  (TIMER2, TIMER_CH_0, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER2, TIMER_CH_0, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER2, TIMER_CH_1, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER2, TIMER_CH_1, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER2, TIMER_CH_2, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER2, TIMER_CH_2, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER2, TIMER_CH_3, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER2, TIMER_CH_3, TIMER_OC_SHADOW_DISABLE);

    timer_channel_output_mode_config  (TIMER3, TIMER_CH_2, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER3, TIMER_CH_2, TIMER_OC_SHADOW_DISABLE);
    timer_channel_output_mode_config  (TIMER3, TIMER_CH_3, TIMER_OC_MODE_PWM0);
    timer_channel_output_shadow_config(TIMER3, TIMER_CH_3, TIMER_OC_SHADOW_DISABLE);

    // enable timers
    timer_enable(TIMER0);
    timer_enable(TIMER2);
    timer_enable(TIMER3);
}

void tim5_init()
{
    rcu_periph_clock_enable(RCU_TIMER5);
    
    timer_deinit(TIMER5);

    // timer configuration
    timer_struct_para_init(&timer_para);
    timer_para.prescaler         = 9;
    timer_para.alignedmode       = TIMER_COUNTER_EDGE;
    timer_para.counterdirection  = TIMER_COUNTER_UP;
    timer_para.period            = (SystemCoreClock / 10000) - 1;    // 1Khz tickrate
    timer_para.clockdivision     = TIMER_CKDIV_DIV1;
    timer_init(TIMER5, &timer_para);

    // interrupt configuration
    timer_interrupt_enable(TIMER5, TIMER_INT_UP);

    // start counting
    timer_enable(TIMER5);
    eclic_irq_enable(TIMER5_IRQn, 15, 0);
}