/*
 * $Id: shiftreg.c 497 2021-07-23 03:42:57Z true $
 * begin 202107xx true
 *
 * simple shift register management via SPI
 */

#include "driver/shiftreg.h"
#include "driver/gd32v_hw_spi.h"


uint8_t sipo_data[SIPO_COUNT];


void shiftreg_update()
{
    uint32_t old_psc;
    
    // run as fast as possible - we're always within spec at our operating speeds
    old_psc = gd32v_spi_get_prescale(SIPO_SPI_DEV);
    gd32v_spi_set_prescale(SIPO_SPI_DEV, SPI_PSC_2);
    gd32v_spi_set_polarity(SIPO_SPI_DEV, 0);

    // update shiftreg
    spi_xfer(SIPO_SPI_DEV, 0, sipo_data, SIPO_COUNT);
    gpio_bit_set(SIPO_LAT_PORT, SIPO_LAT_PIN);
    __asm("nop");
    gpio_bit_reset(SIPO_LAT_PORT, SIPO_LAT_PIN);

    // set SPI back to old speed and polarity
    gd32v_spi_set_prescale(SIPO_SPI_DEV, old_psc);
    gd32v_spi_set_polarity(SIPO_SPI_DEV, 1);
}

void shiftreg_oe(uint8_t enable)
{
    gpio_bit_write(SIPO_OE_PORT, SIPO_OE_PIN, !enable);
}

void shiftreg0_auxpwr(uint8_t enable)
{
    sipo_data[0] &= ~SIPO0_AUXPWR_BITS;
    sipo_data[0] |= enable ? (1 << SIPO0_AUXPWR_SHIFT) : 0;
}

void shiftreg0_edge(uint8_t edgebits)
{
    sipo_data[0] &= ~SIPO0_EDGE_BITS;
    sipo_data[0] |= edgebits << SIPO0_EDGE_SHIFT;
}

void shiftreg0_eyes(uint8_t eyesbits)
{
    sipo_data[0] &= ~SIPO0_EYES_BITS;
    sipo_data[0] |= eyesbits << SIPO0_EYES_SHIFT;
}