/*
 * $Id: fw_app.c 500 2021-08-08 19:43:38Z true $
 * begin 20210618 true
 * 
 * dc29 minibadge
 * basically a port of 8051 code from the DC27 GAT Nametag
 * written in 2021 by true
 * of the whiskey pirates
 * 
 * yeah, the code quality is bad. who gives a fuck. had no time.
 * 
 * known bugs:
 * - sometimes when looking at about screens, particularly temperature, the display stops working
 * - i2c0 and uart1 errors can pile up - uart can be fixed with faster clock, i2c0 is kinda OK right now...
 * - too much division; it takes too long, maybe causes DMA or interrupt to stall? contributes to i2c0/uart1 errors
 * - timeout should only work at main menu, but is working on other menus
 * 
 * todo's:
 * - add various light programs; see the rgbled_x.c files for what I want to do
 * - add gamma correction to LEDs to make low levels work better
 * - replace sleep timer check with adxl interrupt (like wakeup) and checking every second if set
 * - add hysteresis in the rotation so it doesn't look so juddery when idle
 * - add Z axis to rotation to disregard rotation or bias torward upright as badge becomes flat
 * - fix i2c comms to not get stuck in loop if, for example, pins are shorted; add a timeout
 * - fix rendering offscreen causing crashes (out of bounds writing / corruption?)
 * 
 * rough version history:
 * - 0.29.02 first public release
 * - 0.29.03 fixed phantom button press by fixing uart handling, adding idle check (should've done to begin with)
 * - 0.29.04 try to fix freezes / startup issues related to oled, slow framerate to 25fps, fixed character rotation framebuffer corruption
 * 
 */


#include "fw_app.h"

// display shit
#include "disp/draw_ssd1306.h"

// sensor shit
#include "sensor/adxl.h"
#include "sensor/lightlevel.h"
#include "sensor/system.h"

// rgbled shit
#include "led/rgbled.h"

// menus and UI
#include "ui/menu.h"
#include "ui/btn.h"

// other shit
#include "driver/gd32v_hw_adc.h"
#include "driver/gd32v_hw_clk.h"
#include "driver/gd32v_hw_hwid.h"
#include "driver/gd32v_hw_tim.h"
#include "driver/shiftreg.h"
#include "driver/w25q_flash.h"
#include "math/rand16.h"

// user's preferences
#include "user_config.h"


// config
#define MENU_TIMEOUT_TO_NAMETAG		60		// seconds
#define ACCEL_STILL_VARIATION		15		// amount of movement needed to keep awake
#define ACCEL_WAKEUP_THRESH			13		// amount of movement needed to wake up

#define OLED_RETRY_COUNT			4


// obvious
const    uint8_t  vers[] = "0.29.04";
volatile uint32_t rst_src;

// fw status bits
uint8_t fw_status;

// timing
uint16_t tick_ctr = 0;
uint8_t missed = 0;

uint8_t cpu_use = 0;
uint8_t cpu_max = 0;

uint32_t uptime = 0;
uint16_t uptime_hour;
uint8_t  uptime_min;
uint8_t  uptime_sec;

uint32_t idle_time_menu;
uint32_t idle_time_still;
uint8_t  idle_go_sleep;

uint16_t batt_level;
uint8_t  batt_volt;
uint8_t  batt_mv;
uint8_t  batt_is_low = 0;

uint8_t  temp_degc;
uint8_t  temp_degc_decimal;

uint8_t  oled_startup;

uint8_t  run_ledprog = 0;

// adc
uint16_t adc_result[ADC_COUNT];


  // non-timing critical code after oled update
void fw_oled_update_done()
{
	int16_t a;
	int8_t rot;

	// reset oled callback, clear screen, and set default pixel mode and font size
	oled.callback = 0;
	if (!(menu->flags & MENU_FLAG_NO_AUTOCLS)) {
		ssd1306_cls(&oled);
	}
	ssd1306fb_set_color(SSD1306_STATE_SET_PIXEL);
	oled.state &= ~SSD1306_STATE_STR_HALFWIDTH;


	// orientation / flipping flags
	rot = adxl345_get_rotation(&adxl);

	if ((rot > (96+4)) || (rot < (32-4))) {
		sysflags &= ~SYS_OLED_ROTATE_X;
	} else if ((rot > (32+4)) && (rot < (96-4))) {
		sysflags |=  SYS_OLED_ROTATE_X;
	}

	if ((rot > (64+4)) && (rot < 124)) {
		sysflags &= ~SYS_OLED_ROTATE_Y;
	} else if ((rot > 4) && (rot < (64-4))) {
		sysflags |=  SYS_OLED_ROTATE_Y;
	}

	if ((rot < 21) || (rot > (64 + 24))) {
		sysflags &= ~SYS_OLED_REVERSE_CHARS;
	} else if (rot > 24 && rot < (64 + 21)){
		sysflags |=  SYS_OLED_REVERSE_CHARS;
	}


#ifdef MENU_TIMEOUT_TO_NAMETAG
	// root menu idle counting
	if (menu == &menu_0) {
		if (!idle_time_menu) {
			idle_time_menu = uptime;
		} else if ((uptime - idle_time_menu) >= MENU_TIMEOUT_TO_NAMETAG) {
			// been at the root menu too long.
			// return to nametag
			menu_stop(0);
			idle_time_menu = 0;
		}
	} else {
		idle_time_menu = 0;
	}
#endif
	
	// do menu operations
	menu_tick();

	// calculate CPU usage
	// note: we just consider missed ticks a percent.
	// it's actually more like a tenth. fuck it.
	a = missed;

	if (uconf.framemod == UCONF_FRAMERATE_HALF) {
		// the above calculation is tied to framerate.
		// so this will compensate for half framerate mode...
		a >>= 1;
	}

	if (a > cpu_max) {
		cpu_max = a;
	}
	cpu_use = a;


	// reset missed interrupt counter
	missed = 0;
}

static void fw_gpio_init()
{
	// alternate function remapping
	rcu_periph_clock_enable(RCU_AF);
	gpio_pin_remap_config(GPIO_SWJ_NONJTRST_REMAP, ENABLE);

	// PC13 is LED enable with external pullup
	gpio_bit_set(GPIOC, GPIO_PIN_13);
	gpio_init(GPIOC, GPIO_MODE_OUT_OD, GPIO_OSPEED_2MHZ, GPIO_PIN_13);

	// PC14 is !ADXL_CS, PC15 is !FLASH_CS, which are active low outputs
	gpio_bit_set(GPIOC, GPIO_PIN_14 | GPIO_PIN_15);
	gpio_init(GPIOC, GPIO_MODE_OUT_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_14 | GPIO_PIN_15);

	// PA0 is ADXL interrupt input pin, active low
	gpio_init(GPIOA, GPIO_MODE_IN_FLOATING, GPIO_OSPEED_50MHZ, GPIO_PIN_0);

	// PA1 is red_led_k, PA2 is red_led_a. both are normally push/pull but sometimes PA2 is configured for adc operation
	gpio_bit_set(GPIOA, GPIO_PIN_1 | GPIO_PIN_2);
	gpio_init(GPIOA, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, GPIO_PIN_1 | GPIO_PIN_2);
	
	// PA3 is usart1_rx which is used for button data
	gpio_init(GPIOA, GPIO_MODE_IPU, GPIO_OSPEED_50MHZ, GPIO_PIN_3);
	// PA4 is vr1_sw which is used for battery level monitoring
	gpio_init(GPIOA, GPIO_MODE_AIN, GPIO_OSPEED_50MHZ, GPIO_PIN_4);

	// PA5 is spi0_clk, PA7 is spi0_mosi
	gpio_init(GPIOA, GPIO_MODE_AF_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_5 | GPIO_PIN_7);
	// PA6 is spi0_miso
	gpio_init(GPIOA, GPIO_MODE_IPU, GPIO_OSPEED_50MHZ, GPIO_PIN_6);

	// PB0 is t2c2_pwm, PB1 is t2c3_pwm
	gpio_init(GPIOB, GPIO_MODE_AF_PP, GPIO_OSPEED_2MHZ, GPIO_PIN_0 | GPIO_PIN_1);

	// PB2 is boot1 pre-boot, but is the shift register latch output in application; is active high; has external pulldown
	gpio_bit_reset(GPIOB, GPIO_PIN_2);
	gpio_init(GPIOB, GPIO_MODE_OUT_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_2);

	// PB10 is i2c1_scl, PB11 is i2c1_sda, PB6 is i2c0_scl, PB7 is i2c0_sda; has external pullup
	gpio_init(GPIOB, GPIO_MODE_AF_OD, GPIO_OSPEED_50MHZ, GPIO_PIN_6 | GPIO_PIN_7 | GPIO_PIN_10 | GPIO_PIN_11);

	// PB12 is spi1_nss, PB13 is SPI1_SCK, PB15 is SPI1_MOSI
	gpio_init(GPIOB, GPIO_MODE_AF_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_12 | GPIO_PIN_13 | GPIO_PIN_15);
	// PB14 is spi1_miso
	gpio_init(GPIOB, GPIO_MODE_IPU, GPIO_OSPEED_50MHZ, GPIO_PIN_14);

	// PA8-PA10 is t0c0_pwm-t0c2_pwm
	gpio_init(GPIOA, GPIO_MODE_AF_PP, GPIO_OSPEED_50MHZ, GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10);
	
	// PB4 is t2c0_pwm, PB5 is t2c1_pwm, PB8 is t3c2_pwm, PB9 is t3c3_pwm
	gpio_init(GPIOB, GPIO_MODE_AF_PP, GPIO_OSPEED_2MHZ, GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_8 | GPIO_PIN_9);
	// t2c0_pwm and t2c1_pwm need remapped
	gpio_pin_remap_config(GPIO_TIMER2_PARTIAL_REMAP, ENABLE);
}

void fw_init(uint8_t initial)
{
    uint32_t i;

	// load our reset source for checking, then reset it
	rst_src = RCU_RSTSCK & 0xfc000000;
	RCU_RSTSCK |= RCU_RSTSCK_RSTFC;
	// we can also analyze wakeup / standby event later
	PMU_CTL |= (PMU_CTL_WURST | PMU_CTL_STBRST);
    
    // set system frequency clock to main operating speed
    SystemInit();

	// enable debug while asleep; we're hackers after all
	dbg_low_power_enable(DBG_LOW_POWER_SLEEP | DBG_LOW_POWER_DEEPSLEEP);
	    
    // configure our GPIOs
	rcu_periph_clock_enable(RCU_GPIOA);
    rcu_periph_clock_enable(RCU_GPIOB);
    rcu_periph_clock_enable(RCU_GPIOC);
    fw_gpio_init();

	// clear RGBLED colors
	for (i = 0; i < RGB_EDGE_COUNT; i++) {
		hsv_edge[i].h = 0x00;
		hsv_edge[i].s = 0x00;
		hsv_edge[i].v = 0x00;
	}
	for (i = 0; i < RGB_EYES_COUNT; i++) {
		hsv_eyes[i].h = 0x00;
		hsv_eyes[i].s = 0x00;
		hsv_eyes[i].v = 0x00;
	}

	// dma is used with i2c and other peripherals
    rcu_periph_clock_enable(RCU_DMA0);

	// buttons
	btn_init();

	// spi init for spi devices
	rcu_periph_clock_enable(RCU_SPI0);
	gd32v_spi_init(SPI0, SPI_PSC_8);	// /4 = 6MHz SPI at 24MHz. /8 = 4.5MHz at 36MHz. address this when adding USB support
	
	// shiftreg
	sipo_data[0] = SIPO0_AUXPWR_BITS | SIPO0_EDGE_BITS;		// turn on OLED, turn off all RGBLED
	shiftreg_update();
	shiftreg_oe(1);
	
	// accelerometer
	adxl345_init();
	adxl345_tick();		// get some values for randomness init

	// rand16
	rand16_init(0x1337 - ((adxl.x | adxl.y) ^ adxl.z));
	
	// user config
	uconf_load();
	lsens_set_lo_threshold(uconf.lsens_lo_thresh);
	
	// LED timers
	tim_rgb_init();
	
	// others
	hwid_init();

	// enable interrupts
    eclic_global_interrupt_enable();
	eclic_set_nlbits(ECLIC_GROUP_LEVEL4_PRIO0);

	// adc
	rcu_periph_clock_enable(RCU_ADC0);
	gd32v_adc_init();

	// light pulse
	// if we have woken up, pulse the LED not for any light but to reset the state machine
	lsens_pulse(initial ? 10 : 1);
	
	// oled
	for (i = 0; i < 12000; i++) { asm("nop"); }		// ensure we wait at least 1ms after powering oled
	oled_startup = 0;
	i2c_bus_reset(GPIOB, GPIO_PIN_7, GPIO_PIN_6);
	i2c_init();
	ssd1306fb_set_target(&oled);
	ssd1306_init(1);	// we'll try to init later too, since sometimes they fail	

	// start up the menu
	// lol. yes, this "starts" the "menu" in nametag mode
	menu_stop(0);

	// configure timer5 (main firmware tick)
	tim5_init();
}

/*
 * timing-sensitive, direct-in-ISR tasks.
 */
uint8_t edge_examples[][8] = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},	// 00 for 00 solid color: always on
	{0x00, 0x01, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00},	// 01 for 00 solid color: flashing
	{0x00, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x80},	// 02 for 01 flicker: favcolor candle mode
	{0x0a, 0x52, 0x06, 0x00, 0x00, 0x00, 0x00, 0xb0},	// 03 for 01 flicker: fast randomcolor circle chase with fadeout
	{0x0a, 0x56, 0x08, 0x00, 0x00, 0x00, 0x00, 0xb0},	// 04 for 01 flicker: fast randomcolor randompos with fadeout
	{0x16, 0x72, 0x1d, 0x00, 0x00, 0x00, 0x00, 0x80},	// 05 for 01 flicker: slow randomcolor cross with fadeout
	{0x4f, 0x17, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00},	// 06 for 02 circles: altcolor opposing with fadeout
	{0x86, 0x02, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00},	// 07 for 03 waving: typical
	{0x03, 0x10, 0x04, 0x00, 0x00, 0x00, 0xff, 0xa0},	// 08 for 04 rainbow: nice slow full rgb shown at once
	{0x04, 0x60, 0x00, 0x00, 0x00, 0x00, 0xff, 0xa0},	// 08 for 04 rainbow: taste the rainbow
	{0x00, 0x05, 0x08, 0x00, 0x00, 0x44, 0x44, 0xff}
};
void fw_tick_critical()
{
	/*** timing counter ***/
	tick_ctr++;
	if (tick_ctr == 1000) {
		tick_ctr = 0;
		uptime++;
	}

	if (fw_status & FW_STATUS_LOOP_BUSY) {
		missed++;
	}


	/*** rgbled driver ***/
	// we run this right away as any variation will create flicker
	if (uconf.flags & UCONF_FLAGS_LEDS_ENABLE) {
		rgb_eyes_update(tick_ctr & 1);
		rgb_edge_update(tick_ctr % 5);
	} else {
		rgb_eyes_update(0xff);
		rgb_edge_update(0xff);
	}


	/*** accelerometer ***/ 		// ~125Hz, after smoothing 62.5Hz usable data
	// update
	if ((tick_ctr % 8) == 0) {
		adxl345_tick();
	}


	/*** adc (buttons, temperature, ...) ***/
	if (((uptime % 300) == 0) && (tick_ctr < 100)) {
		if (tick_ctr == 5) {
			adc_calibration_enable(ADC0);
		}
	} else {
		gd32v_adc_tick();
	}


	/*** buttons ***/	
	if (tick_ctr & 1) {
		btn_tick();
	}


	/*** rgbled programs ***/
	// verify
	if (uconf.ledprog_edge_idx > (sizeof(edge_pgm) / sizeof(edge_pgm[0]))) {
		uconf.ledprog_edge_idx = 0;	
	}
	if (uconf.ledprog_eyes_idx > (sizeof(eyes_pgm) / sizeof(eyes_pgm[0]))) {
		uconf.ledprog_eyes_idx = 0;	
	}

	// run LED programs at 200hz
	if (uptime || (!uptime && (tick_ctr > 150))) {
		if (tick_ctr % 5 == 0) {
			run_ledprog = 1;
		}
	}
	
	/* autosleep */
	if (uconf.sleep_timeout) {
		if (tick_ctr % 50 == 0) {		
			idle_time_still++;
			
			if (adxl345_movement() > ACCEL_STILL_VARIATION) {
				idle_time_still = 0;
				idle_go_sleep = 0;
			}		

			if (idle_time_still >= (uconf.sleep_timeout * 20)) {
				idle_go_sleep = 1;
			}
		}
	}
}

/*
 * less timing sensitive, yet still timed tasks.
 */
uint32_t oled_retry;
void fw_tick_mainloop()
{
	fw_status |= FW_STATUS_LOOP_BUSY;



	/* per second updating */
	if (tick_ctr == 0) {
		// uptime
		uptime_hour = (uint16_t)(uptime / 3600);
		uptime_min  = (uint8_t)((uptime /   60) % 60);
		uptime_sec  = (uint8_t)((uptime       ) % 60);
	}
	if (tick_ctr == 4) {
		// battery level
		uint16_t batt_level = (adc_get_val_avg(1) * 8);
		batt_level += 950;	// rough hard-coded offset
		batt_volt = batt_level / 10000;
		batt_mv   = (batt_level / 100) % 100;
	}
	if (tick_ctr == 8) {
		// temperature
		temp_degc         = adc_get_temp_degc_x10() / 10;
		temp_degc_decimal = adc_get_temp_degc_x10() % 10;
	}


	
	/*** some OLEDs have dramatically bad startup delay  ***/
	/***  try to re-init every second for first 10 secs ***/
	if ((oled_startup < OLED_RETRY_COUNT) && (oled_retry != uptime) && uptime) {
		ssd1306_init(0);
		oled_startup++;
		oled_retry = uptime;

		// detect errors only after startup - dc29 quick fix for white OLEDs
		// todo: make this a better fix
		if (oled_startup == OLED_RETRY_COUNT) {
			i2c_interrupt_enable(I2C0, I2C_INT_ERR);
		}
	}
	if (i2c_reset) {
		// for some reason the i2c bus has been reset; re-initialize OLED
		i2c_reset = 0;
		oled_startup = 0;
		ssd1306_init(0);
	}
	


	/*** button callback processing ***/
	if (tick_ctr & 1) {
		btn_callback();
	}



	/*** oled ***/
	if (oled.callback && !(oled.state & SSD1306_STATE_BUSY)) {
		if (ssd1306_cb_get()) {
			oled.callback();
		}
	}

	// update
	if ((tick_ctr % uconf.framemod) == 0) {		// 64 == 16fps, gives us some good cycles to update UI
		// only update this frame if we're not in the middle of starting up
		if (!(oled_startup && !tick_ctr)) {
			// process other tasks
			if (oled.state & SSD1306_STATE_INITIALIZED) {
				oled.callback = fw_oled_update_done;
			}
			ssd1306_update();
		}
	}



	/*** status LED ***/
	uint16_t batt = adc_get_val_avg(ADC_CHAN_BATTERY);
	if (batt < 2250) {
		// flash when below ~1.9V, or 0.95V/cell
		batt_is_low = 1;
	} else if (batt > 2650) {
		// at >1.1V/cell, restore battery low state and stop flashing
		batt_is_low = 0;
	}

	if (batt_is_low && ((uptime & 1) == 0) && (tick_ctr < 20)) {
		lsens_pulse(40);
	}



	/*** led programs ***/
	if (run_ledprog) {
		run_ledprog = 0;
		if (edge_pgm[uconf.ledprog_edge_idx].prog) {
			edge_pgm[uconf.ledprog_edge_idx].prog(uconf.ledprog_edge_data[uconf.ledprog_edge_idx], tick_ctr);
		}
		if (eyes_pgm[uconf.ledprog_eyes_idx].prog) {
			eyes_pgm[uconf.ledprog_eyes_idx].prog(uconf.ledprog_eyes_data[uconf.ledprog_eyes_idx], tick_ctr);
		}
	}



	/*** wrapping up ***/
	rand16();

	if (idle_go_sleep) {
		// disable interrupts
		eclic_global_interrupt_disable();
		timer_disable(TIMER5);

		// todo: figure out why deepsleep mode isn't working anymore. some issue with exti?
		// for now we're using standby which has higher power consumption to wake up

		// reset state
		idle_go_sleep = 0;
		idle_time_still = 0;

		// main lights out
		sipo_data[0] = SIPO0_EDGE_BITS;
		shiftreg_update();

		// low batt light out
		gpio_bit_set(GPIOA, GPIO_PIN_1 | GPIO_PIN_2);
		gpio_init(GPIOA, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, GPIO_PIN_1 | GPIO_PIN_2);

		// configure our wakeup interrupt for deepsleep
		//gpio_exti_source_select(GPIO_PORT_SOURCE_GPIOA, GPIO_PIN_SOURCE_0);
		
		// event-based
		//exti_init(EXTI_0, EXTI_EVENT, EXTI_TRIG_RISING);
		//exti_flag_clear(EXTI_0);
		//exti_event_enable(EXTI_0);
		
		// interrupt-based
		//exti_init(EXTI_0, EXTI_INTERRUPT, EXTI_TRIG_RISING);
		//exti_interrupt_flag_clear(EXTI_0);
		//exti_interrupt_enable(EXTI_0);
		//eclic_irq_enable(EXTI0_IRQn, 15, 0);

		// configure accelerometer interrupt
		adxl345_set_reg8(ADXL345_REG_THRESH_ACT, ACCEL_WAKEUP_THRESH);
		adxl345_set_reg8(ADXL345_REG_ACT_INACT_CTL,
							ADXL345_CTL_ACT_AC |
							ADXL345_CTL_ACT_X_ENA |
							ADXL345_CTL_ACT_Y_ENA |
							ADXL345_CTL_ACT_Z_ENA);
		adxl345_get_interrupt_flag();		// reading resets flags, req'd before enabling
		adxl345_set_interrupt(0, ADXL345_INT_ACTIVITY);

		// go to sleep
		rcu_periph_clock_enable(RCU_PMU);
		//// standby mode (long wakeup time / power consumption
		pmu_wakeup_pin_enable();
		pmu_to_standbymode(WFE_CMD);
		//// deepsleep mode (higher idle consumption but better overall with lower wakeup time)
		//pmu_to_deepsleepmode(PMU_LDO_LOWPOWER, WFE_CMD);

		// we've woken up, so soft reset
		//eclic_irq_disable(EXTI0_IRQn);
		fw_init(0);
	}


	fw_status &= ~FW_STATUS_LOOP_BUSY;
}

void fw_adc_complete_cb()
{
	// do shit with the light sensor once the ADC is done reading
	lsens_tick();
}