/*
 * $Id: rgbled.c 499 2021-07-26 05:24:02Z true $
 * begin 20190505 true
 */

#include "led/rgbled.h"

#include "driver/gd32v_hw_tim.h"
#include "driver/shiftreg.h"

#include "math/scale.h"

#include "sensor/lightlevel.h"



color_rgb rgb[2];				// only max 2 lit at a time in matrix

color_hsv hsv_edge[RGB_EDGE_COUNT];
color_hsv hsv_eyes[RGB_EYES_COUNT];

static color_rgb rgb_zero[2] = {{0, 0, 0}, {0, 0, 0}};

       uint16_t eyes_val_lsens = 0;
static uint16_t eyes_val_lsens_lookup[16] = {
	0xfffe, 0xfffe, 0xfffe, 0xeff0,
	0xdff0, 0xbff0, 0x8ff0, 0x5ff0,
	0x4ff0, 0x3ef0, 0x2df0, 0x1ef0,
	0x1af0, 0x17f0, 0x14f0, 0x12f0
};


void rgb_eyes_update(uint8_t idx)
{
	int i;

	// disable outputs
	timer_primary_output_config(TIMER0, DISABLE);

	// LEDs disabled?
	if (idx == 0xff) {
		shiftreg0_eyes(0);
		shiftreg_update();
		return;	
	}

	// convert values
	hsv2rgb(hsv_eyes[idx].h, hsv_eyes[idx].s, hsv_eyes[idx].v, &rgb[0].r, &rgb[0].g, &rgb[0].b);

	// figure out scaling factor
	i = lsens_get_hi();
	if (i > 15) i = 15;
	if (!eyes_val_lsens) {
		// get initial	
		// for some reason this isn't working even though I think it should
		// as I delay execution of LED programs until after we've retrieved
		// ADC data and processed light sensor information
		// so for now just hardcode halfway
		// eyes_val_lsens = eyes_val_lsens_lookup[i];
		eyes_val_lsens = eyes_val_lsens_lookup[7];
	} else {
		// adjust scaled value max
		if (eyes_val_lsens > eyes_val_lsens_lookup[i]) {
			eyes_val_lsens--;
		} else {
			eyes_val_lsens++;
		}
	}

	// force update event, then stop timer
	timer_event_software_generate(TIMER0, TIMER_EVENT_SRC_UPG);
	timer_disable(TIMER0);

	// select new output
	shiftreg0_eyes(idx + 1);
	shiftreg_update();

	// reset timer count
	TIMER_CNT(TIMER0) = 0;

	// update PWMs
	i = eyes_val_lsens >> 6;
	TIMER_CH0CV(TIMER0) = u16_scale(rgb[0].r, 0, 255, 0, i);
	TIMER_CH1CV(TIMER0) = u16_scale(rgb[0].g, 0, 255, 0, i);
	TIMER_CH2CV(TIMER0) = u16_scale(rgb[0].b, 0, 255, 0, i);

	// start counter
	timer_primary_output_config(TIMER0, ENABLE);
	timer_enable(TIMER0);
}	

void rgb_edge_setpwm(color_rgb *u)
{
    TIMER_CH0CV(TIMER2) = u[0].r;
    TIMER_CH1CV(TIMER2) = u[0].g;
    TIMER_CH2CV(TIMER2) = u[0].b;
    TIMER_CH3CV(TIMER2) = u[1].r;

    TIMER_CH2CV(TIMER3) = u[1].g;
    TIMER_CH3CV(TIMER3) = u[1].b;	
}

void rgb_edge_update(uint8_t idx)
{
	int i, j;

	// set all PWMs to 0
	rgb_edge_setpwm(rgb_zero);

	// LEDs disabled?
	if (idx == 0xff) {
		shiftreg0_edge(0x1f);
		shiftreg_update();
		return;
	}

	// convert values
	j = idx << 1;
	for (i = 0; i < 2; i++) {
		hsv2rgb(hsv_edge[j].h, hsv_edge[j].s, hsv_edge[j].v, &rgb[i].r, &rgb[i].g, &rgb[i].b);
		j++;
	}

	// force update event, then stop counters
	timer_event_software_generate(TIMER2, TIMER_EVENT_SRC_UPG);
	timer_disable(TIMER2);
	timer_event_software_generate(TIMER3, TIMER_EVENT_SRC_UPG);
	timer_disable(TIMER3);
	
	// select new output
	i = 0x1f & ~(1 << idx);
	shiftreg0_edge(i);
	shiftreg_update();

	// reset count
	TIMER_CNT(TIMER2) = 0;
	TIMER_CNT(TIMER3) = 0;

	// update PWMs
	rgb_edge_setpwm(rgb);

	// start counters
	timer_enable(TIMER2);
	timer_enable(TIMER3);
}
