/*
 * $Id: rgbled_eyes.c 500 2021-08-08 19:43:38Z true $
 * begin 20210721 true
 *
 * programs to run to show neat shit on the LEDs
 * programs may be duplicated with minimal changes between eyes and edge
 * this is fine right now - we have the space
 * note also that some programs may change once the LEDs run at higher bit depth
 * 
 * programs I'd like to get done:
 * - things that will do color fades between colors; quickest hack would be implementing rgb2hsv
 */


#include "led/rgbled.h"

#include "sensor/adxl.h"

#include "math/rand16.h"
#include "math/scale.h"
#include "math/sin7.h"

#include "user_config.h"



static uint8_t timeout;


void eyes_solid(uint8_t *a, uint16_t tick)
{
	// 1=bitfield, 2=timeout-set, 4=state, 5=hue, 6=sat, 7=val
	// bitfield:
	// 7: blank / off
	// 3: override altcolor (if bit 1 is not set)
	// 1: override hsv
	// 0: flash

	color_hsv hsv = {0, 0, 0};

	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}

	if (a[1] & 0x80) {
        // override off
		hsv.v = 0;
	} else {
        // color select
        if (a[1] & 0x02) {
            hsv.h = a[5] * 6;
            hsv.s = a[6];
            hsv.v = a[7];
        } else if (a[1] & 0x08) {
            hsv.h = uconf.altcolor_hue * 6;
            hsv.s = uconf.altcolor_sat;
            hsv.v = uconf.altcolor_val;
        } else {
            hsv.h = uconf.favcolor_hue * 6;
            hsv.s = uconf.favcolor_sat;
            hsv.v = uconf.favcolor_val;
        }

        // flash
        if (a[1] & 0x01) {
            a[4] ^= 0x01;
            if (a[4] & 0x01) hsv.v = 0;
        }
    }	

	// update
	hsv_eyes[0].h = hsv_eyes[1].h = hsv.h;
	hsv_eyes[0].s = hsv_eyes[1].s = hsv.s;
	hsv_eyes[0].v = hsv_eyes[1].v = hsv.v;
}

// todo: improve fading smoothness by doing fadeout every callback instead of on timeout
// this can be done once LED bit depth is increased
void eyes_flicker(uint8_t *a, uint16_t tick)
{
	// 0=speed, 1=bitfield, 2=timeout-set, 4=state, 6=max-val, 7=min-val
	// bitfield:
	// 7: min-val is not half range (0-255); if unset min-val is half range (128-255)
	// 6: apply a random hue instead of favcolor
	// 4: fades out all LEDs at speed setting (only if bit 1 SET)
	// 3: same flicker throughout (only if bit 1 NOT SET)
	// 2: make the LED random instead of sequential (only if bit 1 SET)
	// 1: only process one LED per update
	// 0: in single LED mode, reverse direction
	// speed: fadeout speed

	int i;
    color_hsv hsv;
	uint8_t min, max;

	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}
	
	// set constants
	hsv.h = uconf.favcolor_hue * 6;
	hsv.s = uconf.favcolor_sat;

    if (a[6]) max = a[6]; else max = 255;

	// set value limit
	if (a[1] & 0x80) {
		min = a[7];
	} else {
		min = (a[7] >> 1) | 0x80;
	}

	// process and update LEDs
	for (i = 0; i < RGB_EYES_COUNT; i++) {
		// compute new val on first round or if all-same flicker is disabled
		if (!(a[1] & 0x08) || i == 0) {	
			hsv.v = u16_scale((rand16() & 0xff), 0, 255, min, max);
			hsv.v = (uconf.favcolor_val * hsv.v) / 256;			
		}

		// only doing one LED?
		if (a[1] & 0x02) {
			// is it random?
			if (a[1] & 0x04) {
				i = rand16() % RGB_EYES_COUNT;
			} else {
				// nope, sequential
				a[4]++;
				a[4] %= RGB_EYES_COUNT;
				i = a[4] % RGB_EYES_COUNT;
				
				// is it in reverse?
				if (a[1] & 0x01) {
					i = RGB_EYES_COUNT - 1 - i;
				}
			}
		}
		
		// apply
		hsv_eyes[i].h = (a[1] & 0x40) ? (rand16() & 0x600) : hsv.h;
		hsv_eyes[i].s = hsv.s;
		hsv_eyes[i].v = hsv.v;

		// bail if only doing one LED
		if (a[1] & 0x02) {
			// but make sure to fade LEDs if needed
			if (a[1] & 0x10) {
				for (i = 0; i < RGB_EYES_COUNT; i++) {
					if (hsv_eyes[i].v <= a[0]) {
						hsv_eyes[i].v = 0;
					} else {
						hsv_eyes[i].v -= a[0];
					}
				}
			}
			return;
		}
	}
}

uint16_t eyes_altfade_activehue;
uint8_t eyes_altfade_stall;
void eyes_altfade_unused(uint8_t *a, uint16_t tick)
{
    // 0=stall_time, 1=bitfield, 2=timeout-set, 4=work, 5=left-hue, 6=right-hue, 7=sat/val
    // bitfield:
    // 7 use specified hsv instead of favcolors
    // 6 use random hue instead of favcolor (if bit 7 is NOT set)
    // 1 both fade in/out (unset = independent alternating)
    // 0 swap left/right
    // stall_time:
    // a[0][7:4]: fading rate
    // a[0][0:3]: stall time delay
    // sat/val:
    // a[7][7:4]: sat
    // a[7][0:3]: hue
    // note: function is incomplete

    uint8_t inc;

    color_hsv hsv;
    uint8_t idx;
    
    uint8_t state;
    uint8_t iter;

	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}

    state = a[4] & 0x7;
    iter = a[4] >> 4;

    // get color
    hsv.s = ((state >> 2) ? uconf.altcolor_sat : uconf.favcolor_sat) * 6;
    hsv.v = ((state >> 2) ? uconf.altcolor_val : uconf.favcolor_val) * 6;
    
    if (a[1] & 0x80) {
        // specified color
        hsv.h = a[5 + (state >> 2)] * 6;
        hsv.s = u16_scale(a[7] >> 4,   0, 0xf, 0, 0xff);
        hsv.v = u16_scale(a[7] & 0xff, 0, 0xf, 0, 0xff);
    } else if (a[1] & 0x40) {
        // random color
        hsv.h = rand16() & 0x5ff;
    } else {
        // favcolor
        hsv.h = ((state >> 2) ? uconf.altcolor_hue : uconf.favcolor_hue) * 6;
    }

    // swap
    idx = (a[1] & 0x01) ? 1 : 0;

    // perform action
    inc = 0;
    switch (state) {
        case 0:
        case 4: {
            // fade up
            
        }
        case 2:
        case 6: {
            // fade down
        }
        default: {
            // stall wait
        }
    }

    if (inc) {
        state++;
    }

    // save state
    a[4] = (iter << 4) | state;

    // which one is active?
    if (state >= 4) {
        idx++;
        idx &= 0x01;
    }

    // output
    hsv_eyes[idx].h = hsv.h;
    hsv_eyes[idx].s = hsv.s;
    hsv_eyes[idx].v = hsv.v;

    hsv_eyes[(idx+1) & 0x01].v = 0;
}

uint8_t eyes_fading_counter = 0;
uint8_t eyes_fading_sec_ticks = 0;
void eyes_colorpulse(uint8_t *a, uint16_t tick)
{
	// 0=speed, 1=bitfield, 2=timeout-set, 3=state, 4=hue-left, 5=hue-right, 6=sat, 7=val
	// bitfield:
	// 7: apply specified hue instead of favcolor
	// 6: apply a random hue
	// 5: trails desaturate
	// 4: secondary value is half of primary
	// 3: opposing is same hue as primary (unset is alt color hue/sat; is always primary val)
	// 2: enable opposing ball
	// 1: enable fading trails
	// others:
	// a[0][7:4]: fade rate of trails
	// a[0][0:3]: trail faderate divider

	int i;
	uint16_t h, h2;
	uint8_t s, v;
	uint8_t s2;
    uint8_t x;	

	uint8_t trailfade = a[0] >> 4;
    uint8_t traildivf = a[0] & 0xf;
	uint8_t desatfade = a[1] & 0x20;

	// fading
	eyes_fading_counter++;
	if (traildivf) {
        eyes_fading_counter %= traildivf;
    } else {
        eyes_fading_counter = 0;
    }
	if (!eyes_fading_counter && (a[1] & 0x02)) {
		for (i = 0; i < RGB_EYES_COUNT; i++) {
			if (hsv_eyes[i].v <= trailfade) {
				hsv_eyes[i].v = 0;
			} else {
				hsv_eyes[i].v -= trailfade;
			}

			if (desatfade) {
				// fade to white too
				if (hsv_eyes[i].s <= (trailfade >> 1)) {
					hsv_eyes[i].s = 0;
				} else {
					hsv_eyes[i].s -= (trailfade >> 1);
				}
			}
		}
	}

	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}

	uint8_t trail  = a[1] & 0x02;
	uint8_t second = a[1] & 0x04;
	uint8_t scolor = a[1] & 0x08;
	uint8_t secval = a[1] & 0x10;
	uint8_t rndhue = a[1] & 0x40;
	uint8_t sethue = a[1] & 0x80;

	// state
	x = (a[3] & 0xf) + 1;
	x %= RGB_EYES_COUNT;

	// save state
	a[3] = x;

	// colors
	s = s2 = a[6];
	v = a[7];

	if (sethue) {
		h =  a[5] * 6;
        h2 = a[4] * 6;
	} else if (rndhue) {
		h = rand16() & 0x5ff;
        h2 = rand16() & 0x5ff;
	} else {
		h = uconf.favcolor_hue * 6;
		s = uconf.favcolor_sat;
		v = uconf.favcolor_val;
        h2 = uconf.altcolor_hue * 6;
		s2 = uconf.altcolor_sat;
	}

	// secondary
	if (scolor) {
		h2 = h;
		s2 = s;
	}

	// set the next item
	hsv_eyes[x].h = h;
	hsv_eyes[x].s = s;
	hsv_eyes[x].v = v;

	if (second) {
		x++;
        x &= 0x01;
        
        if (secval) v >>= 1;
		
		hsv_eyes[x].h = h2;
		hsv_eyes[x].s = s2;
		hsv_eyes[x].v = v;
	}

	// clear those that are on if trails are not enabled
	if (!trail) {
		for (i = 0; i < RGB_EYES_COUNT; i++) {
			if (i != x && !second) {
				hsv_eyes[i].v = 0;
			}
		}
	}	
}

void eyes_future2(uint8_t *a, uint16_t tick)
{

}

void eyes_rainbow(uint8_t *a, uint16_t tick)
{
	// 0=angle-rate, 1=bitfield, 2=timeout-set, 3=timeout-work, 45=angle-work, 6=sat, 7=val
	// bitfield:
	// 7:4 hue angle division (360 divided by 1-8; all other values mean 1, except 0 which means no offset)
	// 3
	// 2
	// 1
	// 0 direction

	uint16_t *rb_angle = (uint16_t *)&a[4];
	uint16_t angle;
    uint16_t hoffset;
	int i;
	int r;

	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}

	// rainbow hue angle increment
	*rb_angle += (a[0]);
	if (*rb_angle >= 0x600) *rb_angle -= 0x600;
	angle = *rb_angle;

	// hue offset
    i = a[1] >> 4;
    hoffset = 0x300;
    if (i == 0) {
        hoffset = 0;
    } else if (i >= 2 && i <= 8) {
        hoffset = 0x300 / i;
    }

	// apply to LEDs
	for (i = 0; i < RGB_EYES_COUNT; i++) {
		r = (a[1] & 0x01) ? RGB_EYES_COUNT - 1 - i : i;
        hsv_eyes[r].h = angle;
		hsv_eyes[r].s = a[6];
		hsv_eyes[r].v = a[7];

		angle += hoffset;
		if (angle >= 0x600) angle -= 0x600;
	}
}

void eyes_copmode(uint8_t *a, uint16_t tick)
{
	// 0=work, 1=bitfield, 2=timeout-set, 3=timeout-work, 4=work, 56=steps, 7=val
	// bitfield:
	// 7 sync to tick 0
	// 6 skip pattern 2
	// 5 skip pattern 1
	// 4 don't override sat/val to maximums (not implemented)
	// 3 use different colors on either half during pattern 2
	// 2 use different colors on either half during pattern 1
	// 1 repeat subpattern 1
	// 0 use cop colors instead of pri/alt
	// others:
	// a[5][7:4] amount of time to delay in pattern 2
	// a[5][0:3] amount of times to repeat pattern 2 (solid color alternate)
	// a[6][7:4] amount of strobe pulses in pattern 1
	// a[6][0:3] amount of times to repeat pattern 1 (strobes)
	
	uint8_t pattern, iter;
	uint8_t *seq;
	uint8_t w, x;
	uint8_t b;

	uint8_t cnt[4];

	color_hsv hsv;
	color_hsv hsv2;
	color_hsv pri, sec;

	// synchronize
	if (a[1] & 0x80) {
		if (!tick) {
			a[3] = 0;
            a[1] &= ~0x80;
		} else {
			return;
		}
	}
	
	// timeout
	if (!timeout || (timeout > a[2])) {
		timeout = a[2];
	} else {
		timeout--;
		return;
	}

	// colors to use
	if (a[1] & 0x01) {
		pri.h = 0x000;
		pri.s = 0xff;
		sec.h = 0x400;
		sec.s = 0xff;
	} else {
		pri.h = uconf.favcolor_hue * 6;
		pri.s = uconf.favcolor_sat;
		sec.h = uconf.altcolor_hue * 6;
		sec.s = uconf.altcolor_sat;
	}

	// pattern mode
	pattern  = a[0] >> 6;
	iter     = a[0] & 0x3f;
	seq      = &a[4];

	cnt[0] = a[6] & 0xf;
	cnt[1] = a[6] >> 4;
	cnt[2] = a[5] & 0xf;
	cnt[3] = a[5] >> 4;

	if ((pattern == 0) && (a[1] & 0x20)) pattern++;
	if ((pattern == 1) && (a[1] & 0x40)) pattern++;

	if (pattern > 1) pattern = 0;

	switch (pattern) {
		// strobe
		case 0: {
			w = cnt[1] << 1;
			if (!w) w = 32;
			b = (a[1] & 0x02) ? 0x02 : 0x01;
			x = cnt[0];
			if (!x) x = 16;
			x <<= b;			

			// set main color
			hsv.h = (iter & b) ? sec.h : pri.h;
			hsv.s = (iter & b) ? sec.s : pri.s;
			hsv.v = 0;

			// p1 alternate color mode
			if (a[1] & 0x04) {
				hsv2.h = (iter & b) ? pri.h : sec.h;
				hsv2.s = (iter & b) ? pri.s : sec.s;
			} else {
				hsv2.h = hsv.h;
				hsv2.s = hsv.s;
			}

			// set value if on
			if ((*seq < w) && !(*seq & 1)) {	
				hsv.v = a[7];
			}

			if (*seq > w) {
				// iteration done
				iter++;
				*seq = 0;
				if (iter >= x) {
					// pattern done
					pattern++;
					iter = 0;
				}

				break;
			}

			(*seq)++;
			break;
		}
		
		// solid alternate
		case 1: {
			w = cnt[3] << 1;
			if (!w) w = 32;
			x = cnt[2] << 1;
			if (!x) x = 32;

			// set main color
			hsv.h = (iter & 1) ? sec.h : pri.h;
			hsv.s = (iter & 1) ? sec.s : pri.s;
			hsv.v = 0;

			// p2 alternate color mode
			if (a[1] & 0x08) {
				hsv2.h = (iter & 1) ? pri.h : sec.h;
				hsv2.s = (iter & 1) ? pri.s : sec.s;
			} else {
				hsv2.h = hsv.h;
				hsv2.s = hsv.s;
			}

			// set value
			hsv.v = a[7];

			if (*seq > w) {
				// iteration done
				iter++;
				*seq = 0;
				if (iter >= x) {
					// pattern done
					pattern++;
					iter = 0;
				}

				break;
			}

			(*seq)++;
			break;
		}
	}

	// we only have two patterns
	pattern %= 2;

	// save state
	a[0] = (pattern << 6) | (iter & 0x3f);

	// apply to LEDs
    hsv_eyes[0].h = hsv.h;
    hsv_eyes[0].s = hsv.s;
    hsv_eyes[0].v = hsv.v;
    
    hsv_eyes[1].h = hsv2.h;
    hsv_eyes[1].s = hsv2.s;
    hsv_eyes[1].v = hsv.v;
}


// implemented program table
const LedProgram eyes_pgm[6] = {
	{"Solid Color",		eyes_solid},
	{"Flicker",			eyes_flicker},
	{"Color Pulse", 	eyes_colorpulse},
	{"Future 2",		eyes_future2},
	{"Rainbow",			eyes_rainbow},
	{"Cop Mode",		eyes_copmode},
};