/*
 * $Id: lightlevel.c 499 2021-07-26 05:24:02Z true $
 * begin 20190505 true
 *
 * continuing the tradition of using LEDs as crude light meters
 */

#include "sensor/lightlevel.h"
#include "gd32vf103.h"


#define RECAL_INIT_TICKS	3500	// 3.5s calibration period


static uint16_t init;
static uint16_t timeout;
static uint8_t state;

static uint16_t lo_thresh = 0xffff;	// 0x6a0; // sane value on dev board

static uint16_t lsens_hi_work;		// temporarily used while reading
static uint16_t lsens_hi;			// last ms used to read
static uint16_t lsens_lo;			// raw adc value


void lsens_pulse(uint16_t cnt)
{
	if (init) return;

	// light up led
	gpio_bit_set(LSENS_PORT, LSENS_PIN_A);
	gpio_bit_reset(LSENS_PORT, LSENS_PIN_K);
	gpio_init(LSENS_PORT, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, LSENS_PIN_A | LSENS_PIN_K);

	// this is how long we'll be on for in mainline ticks
	timeout = cnt << 1;
}

uint16_t lsens_get_hi()
{
	return lsens_hi;
}
uint16_t lsens_get_lo()
{
	return lsens_lo;
}
uint16_t lsens_get_lo_threshold()
{
	return lo_thresh;
}
void lsens_set_lo_threshold(uint16_t thresh)
{
	lo_thresh = thresh;
}

void lsens_tick()
{
	uint16_t a;

	// is LED lit?
	if (timeout) {
		timeout--;
		if (!timeout) {
			// turn off LED
			gpio_bit_set(GPIOA, LSENS_PIN_K);
			// reset state
			state = 0;
		}
		return;
	}

	if (lo_thresh == 0xffff) {
		if (state != 0xff) state = 0xfe;
	}

	// ok, we can try to read the light value
	switch (state) {
		case 0: {
			// reverse bias to charge LED
			gpio_init(LSENS_PORT, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, LSENS_PIN_A | LSENS_PIN_K);
			gpio_bit_set(LSENS_PORT, LSENS_PIN_K);
			gpio_bit_reset(LSENS_PORT, LSENS_PIN_A);
			// float the cathode
			gpio_init(LSENS_PORT, GPIO_MODE_AIN, GPIO_OSPEED_2MHZ, LSENS_PIN_K);
			
			// prepare acquisition
			lsens_hi_work = 0;
			state++;
			break;
		}
		case 1: {
			lsens_hi_work++;
			a = adc_get_val_raw(ADC_CHAN_LIGHTSENS);
			
			// have we reached our threshold?
			if (a > lo_thresh) {
				// nope, bail
				break;
			}

			// we're at our threshold
			lsens_hi = lsens_hi_work;
			lsens_lo = a;

			// start over again
			state = 0;
			break;
		}
		case 0xfe: {
			// reverse bias to charge LED
			gpio_init(LSENS_PORT, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, LSENS_PIN_A | LSENS_PIN_K);
			gpio_bit_set(LSENS_PORT, LSENS_PIN_K);
			gpio_bit_reset(LSENS_PORT, LSENS_PIN_A);
			// float the cathode
			gpio_init(LSENS_PORT, GPIO_MODE_AIN, GPIO_OSPEED_2MHZ, LSENS_PIN_K);
			
			init = RECAL_INIT_TICKS;
			state++;
			break;
		}
		case 0xff: {
			init--;
			
			a = adc_get_val_raw(ADC_CHAN_LIGHTSENS);
			if ((lo_thresh > a) && a) {
				lo_thresh = a;
			}

			if (!init) {
				// increase thershold to be above noise a bit
				lo_thresh += 50;
				state = 0;
			}
		}
	}
}

