/*
 * $Id: menu_entry_6.c 500 2021-08-08 19:43:38Z true $
 * begin 20190527 true
 *
 * about menu functions
 */

#include "ui/btn.h"
#include "ui/menu.h"

#include "driver/gd32v_hw_i2c.h"
#include "driver/gd32v_hw_hwid.h"

#include "led/rgbled.h"
#include "sensor/adxl.h"
#include "sensor/system.h"
#include "sensor/lightlevel.h"

#include "user_config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "riscv_encoding.h"


uint8_t font_index = 0;
uint8_t font_glyph = 0;

uint8_t sn_byte;



void menu_6_font_index(int8_t dir)
{
	font_index += dir;
	if (font_index >= (sizeof(font_table) / sizeof(font_table[0]))) {
		if (dir > 0) {
			font_index = 0;
		} else {
			font_index = (sizeof(font_table) / sizeof(font_table[0])) - 1;
		}
	}
}

void menu_6_font_glyph(int8_t dir)
{
	font_glyph += dir;
	if (font_glyph > font_table[font_index].font[FONT_CHAR_NUM_POS] - 1) {
		if (dir > 0) {
			font_glyph = 0;
		} else {
			font_glyph = font_table[font_index].font[FONT_CHAR_NUM_POS] - 1;
		}
	}
}

void menu_6_font_next()
{
	if (btn[1].held) {
		menu_6_font_index(1);
	} else {
		menu_6_font_glyph(1);
	}
}

void menu_6_font_prev()
{
	if (btn[2].held) {
		menu_6_font_index(-1);
	} else {
		menu_6_font_glyph(-1);
	}
}

void menu_6_accel_reset()
{
	movement_worst = 0;
	
	/*
	this is where I was testing bootloader entry.
	best I can do is cause a reset. may not even
	be possible to jump to this bootloader from user code
	...after reading online it looks like nope...
	*/
	/*
	void (*bl)() = (void(*)())(*((uint32_t *)0x1fffb000));
	
	eclic_global_interrupt_disable();
	write_csr(0x7d0, 0x0);		// mmisc_ctl
	write_csr(0x7c3, 0x0);		// mnvec
	write_csr(0x7ec, 0x1fffb000);		// mtvt2
	write_csr(0x307, 0x1fffb000);		// mtvt
	
	bl();
	*/
}

void menu_6_btn_use()
{
	btn[1].push_cb = &menu_6_font_prev;
	btn[1].held_cb = &menu_6_font_prev;
	btn[1].hold_time = 8;
	btn[1].hold_retrig = 61;

	btn[2].push_cb = &menu_6_font_next;
	btn[2].held_cb = &menu_6_font_next;
	btn[2].hold_time = 8;
	btn[2].hold_retrig = 61;
}

void menu_6_disp(uint8_t idx)
{
	uint8_t i;
	char txt[30];

	const uint8_t led_pos[4][2] = {
		{10, 1},
		{82, 1},
		{10, 17},
		{82, 17},		
	};

	if (!sn_byte) {
		if (BTN_DATA_VALID && BTN_DATA_SN_BYTE) {
			sn_byte = BTN_DATA_SN_BYTE;
		}
	}

	// which item selected?
	switch (idx) {
		case 0: {
			ssd1306fb_set_cursor(11, 0);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "WP DC29", 1);
			ssd1306fb_set_cursor(13, 11);
			ssd1306fb_draw_str(font_Dialog_plain_8, "minibadge", 1);
			ssd1306fb_set_cursor(80, 0);
			ssd1306fb_draw_str(font_Dialog_plain_8, "by true", 1);
			
			if (sn_byte) {
				sprintf(txt, "#%d", sn_byte);
				i = ssd1306fb_get_str_width(font_DejaVu_Sans_Mono_Bold_11, txt, strlen(txt), 0);
				ssd1306fb_set_cursor(97 - (i >> 1), 10);
				ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, txt, 1);
			}			

			sprintf(txt, "%s", vers);
			ssd1306fb_set_cursor(11, 24);
			ssd1306fb_draw_str(font_Dialog_plain_8, "v", 0);
			ssd1306fb_set_cursor(15, 24);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);
			ssd1306fb_set_cursor(54, 24);
			ssd1306fb_draw_str(font_Dialog_plain_8, "@WhiskeyHackers", 0);
			break;
		}
		case 1: {
			ssd1306fb_set_cursor(54, 4);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "LED", 1);
			ssd1306fb_set_cursor(54, 15);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "1-4", 1);

			for (i = 0; i < 4; i++) {
				ssd1306fb_set_cursor(led_pos[i][0], led_pos[i][1]);
				sprintf(txt, "R%d h%03X", i + 1, hsv_edge[edge_map[i + 0]].h);
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
				sprintf(txt, "s%02X v%02X", hsv_edge[edge_map[i + 0]].s, hsv_edge[edge_map[i + 0]].v);
				oled.cursor_x = led_pos[i][0];
				oled.cursor_y += 7;
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			}
			break;
		}
		case 2: {
			ssd1306fb_set_cursor(54, 4);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "LED", 1);
			ssd1306fb_set_cursor(54, 15);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "5-8", 1);

			for (i = 0; i < 4; i++) {
				ssd1306fb_set_cursor(led_pos[i][0], led_pos[i][1]);
				sprintf(txt, "R%d h%03X", i + 5, hsv_edge[edge_map[i + 4]].h);
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
				sprintf(txt, "s%02X v%02X", hsv_edge[edge_map[i + 4]].s, hsv_edge[edge_map[i + 4]].v);
				oled.cursor_x = led_pos[i][0];
				oled.cursor_y += 7;
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			}
			break;
		}
		case 3: {
			ssd1306fb_set_cursor(54, 4);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "LED", 1);
			ssd1306fb_set_cursor(56, 15);
			ssd1306fb_draw_str(font_Dialog_plain_8, "9-10", 1);
			ssd1306fb_set_cursor(55, 22);
			ssd1306fb_draw_str(font_Dialog_plain_8, "Eyes", 1);

			for (i = 0; i < 2; i++) {
				ssd1306fb_set_cursor(led_pos[i][0], led_pos[i][1]);
				sprintf(txt, "%s h%03X", i == 0 ? "R9" : "10", hsv_edge[edge_map[i + 8]].h);
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
				sprintf(txt, "s%02X v%02X", hsv_edge[edge_map[i + 8]].s, hsv_edge[edge_map[i + 8]].v);
				oled.cursor_x = led_pos[i][0];
				oled.cursor_y += 7;
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			}
			for (i = 0; i < 2; i++) {
				ssd1306fb_set_cursor(led_pos[i + 2][0], led_pos[i + 2][1]);
				sprintf(txt, "E%s h%03X", i == 0 ? "L" : "R", hsv_eyes[i].h);
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
				sprintf(txt, "s%02X v%02X", hsv_eyes[i].s, hsv_eyes[i].v);
				oled.cursor_x = led_pos[i][0];
				oled.cursor_y += 7;
				ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			}
			break;
		}
		case 4: {
			ssd1306fb_set_cursor(10, -1);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "Accelerometer", 1);

			sprintf(txt, "X: %3d", adxl.x);
			ssd1306fb_set_cursor(11, 10);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			sprintf(txt, "Y: %3d", adxl.y);
			ssd1306fb_set_cursor(11, 17);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			sprintf(txt, "Z: %3d", adxl.z);
			ssd1306fb_set_cursor(11, 24);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);
			sprintf(txt, "m%i", abs(adxl345_movement()));
			ssd1306fb_set_cursor(104, 10);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);
			sprintf(txt, "w%d", movement_worst);
			ssd1306fb_set_cursor(106, 17);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			ssd1306fb_draw_line(80, 23, 80 + (adxl.x >> 1), 23 + (adxl.y >> 2));

			ssd1306fb_set_color(SSD1306_STATE_INVERT_PIXEL);
			ssd1306fb_draw_circle(80 + (adxl.x >> 1), 23 + (adxl.y >> 2), 2);
			ssd1306fb_set_color(SSD1306_STATE_SET_PIXEL);

			break;
		}
		case 5: {
			sprintf(txt, "CPU Load: %3u%%", cpu_use);
			ssd1306fb_set_cursor(10, -1);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, txt, 1);

			sprintf(txt, "CPU  Max: %3u%%", cpu_max);
			ssd1306fb_set_cursor(10, 9);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, txt, 1);

			ssd1306fb_set_cursor(10, 23);
			sprintf(txt, "Up %01d:%02d:%02d", uptime_hour, uptime_min, uptime_sec);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);
			break;
		}
		case 6: {
			ssd1306fb_set_cursor(10, -1);
			ssd1306fb_draw_str(font_Dialog_plain_8, "Light: ", 1);
			oled.cursor_x = 39;
			sprintf(txt, "%d lo, %02d + %d", lsens_get_lo_threshold(), lsens_get_hi(), lsens_get_lo());
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			ssd1306fb_set_cursor(10, 7);
			ssd1306fb_draw_str(font_Dialog_plain_8, "Temp:        Batt:", 1);
			
			oled.cursor_x = 42;
			sprintf(txt, "%d.%dC", temp_degc, temp_degc_decimal);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);
			
			oled.cursor_x = 98;
			sprintf(txt, "%d.%02dV", batt_volt, batt_mv);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			ssd1306fb_set_cursor(10, 16);
			ssd1306fb_draw_str(font_Dialog_plain_8, "Flash Conf Writes: ", 0);
			sprintf(txt, "%lu", uconf.iter);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);

			ssd1306fb_set_cursor(10, 24);
			ssd1306fb_draw_str(font_Dialog_plain_8, "MCU: ", 1);
			sprintf(txt, "%dK / %dK", hwid_get_flash_size(), hwid_get_sram_size());
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 1);

			break;
		}
		case 7: {
			ssd1306fb_set_cursor(10, -1);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "Error Counters", 1);

			ssd1306fb_set_cursor(10, 10);
			sprintf(txt, "i2c0: %lu  e# %02x", i2c_err_cnt[0], i2c_last_err_reason[0]);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			ssd1306fb_set_cursor(10, 17);
			sprintf(txt, "i2c1: %lu  e# %02x", i2c_err_cnt[1], i2c_last_err_reason[1]);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);
			
			ssd1306fb_set_cursor(10, 24);
			sprintf(txt, "uart1: s%lu  i%lu", btn_short_packet_cnt, btn_invalid_packet_cnt);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			ssd1306fb_set_cursor(74, 10);
			sprintf(txt, "brt-pk: %04x", eyes_val_lsens);
			ssd1306fb_draw_str(font_Dialog_plain_8, txt, 0);

			break;
		}
		case 8: {
			ssd1306fb_set_cursor(10, -1);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, "Font Test", 1);

			sprintf(txt, "Index:%3u", (uint8_t)font_index);
			ssd1306fb_set_cursor(10, 9);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, txt, 1);

			sprintf(txt, "Glyph:%3u", (uint8_t)(font_glyph + 0x20));
			ssd1306fb_set_cursor(10, 19);
			ssd1306fb_draw_str(font_DejaVu_Sans_Mono_Bold_11, txt, 1);

			sprintf(txt, "%c", (font_glyph + 0x20));
			ssd1306fb_draw_rect(83, 3, 116, 28);
			ssd1306fb_set_cursor(85, 5);
			ssd1306fb_draw_str(font_table[font_index].font, txt, 1);
			break;
		}
	}

	// draw extras
	menu_draw_tabs(idx);

	// buttons
	switch (idx) {
		case 0:
		case 1:
		case 2:
		case 3: {
			menu_btn_use_std();
			break;
		}
		case 4: {
			menu_btn_use_std();
			btn[2].push_cb = menu_6_accel_reset;
			break;
		}
		case 8: {
			menu_6_btn_use();
			menu_draw_buttons(MENU_BTNSTYLE_ABOUT, 0x0c);
			break;
		}
		default: {
			menu_btn_use_std();
			menu_draw_buttons(MENU_BTNSTYLE_ABOUT, 0x08);
			break;
		}
	}
}


